@@ -20,6 +20,7 @@
 
 #include "wconfig.h"
 
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
@@ -31,15 +32,12 @@
 #include "window.h"
 #include "defaults.h"
 #include "switchpanel.h"
-#include "funcs.h"
+#include "misc.h"
 #include "xinerama.h"
 
-extern Atom _XA_WM_IGNORE_FOCUS_EVENTS;
 
-#ifdef SHAPE
+#ifdef USE_XSHAPE
 #include <X11/extensions/shape.h>
-
-extern Bool wShapeSupported;
 #endif
 
 struct SwitchPanel {
@@ -50,6 +48,7 @@ struct SwitchPanel {
 	WMArray *icons;
 	WMArray *images;
 	WMArray *windows;
+	WMArray *flags;
 	RImage *bg;
 	int current;
 	int firstVisible;
@@ -64,14 +63,14 @@ struct SwitchPanel {
 	WMColor *white;
 };
 
-extern WPreferences wPreferences;
-
 #define BORDER_SPACE 10
 #define ICON_SIZE 48
 #define ICON_TILE_SIZE 64
 #define LABEL_HEIGHT 25
 #define SCREEN_BORDER_SPACING 2*20
-#define SCROLL_STEPS (ICON_TILE_SIZE/2)
+
+#define ICON_SELECTED (1<<1)
+#define ICON_DIM (1<<2)
 
 static int canReceiveFocus(WWindow *wwin)
 {
@@ -95,17 +94,47 @@ static int canReceiveFocus(WWindow *wwin)
 	return 1;
 }
 
-static void changeImage(WSwitchPanel *panel, int idecks, int selected)
+static Bool sameWindowClass(WWindow *wwin, WWindow *curwin)
 {
-	WMFrame *icon = WMGetFromArray(panel->icons, idecks);
-	RImage *image = WMGetFromArray(panel->images, idecks);
+	if (!wwin->wm_class || !curwin->wm_class)
+		return False;
+	if (strcmp(wwin->wm_class, curwin->wm_class))
+		return False;
+
+	return True;
+}
+
+static void changeImage(WSwitchPanel *panel, int idecks, int selected, Bool dim, Bool force)
+{
+	WMFrame *icon = NULL;
+	RImage *image = NULL;
+	int flags;
+	int desired = 0;
+
+	/* This whole function is a no-op if we aren't drawing the panel */
+	if (!wPreferences.swtileImage)
+		return;
+
+	icon = WMGetFromArray(panel->icons, idecks);
+	image = WMGetFromArray(panel->images, idecks);
+	flags = (int) (uintptr_t) WMGetFromArray(panel->flags, idecks);
+
+	if (selected)
+		desired |= ICON_SELECTED;
+	if (dim)
+		desired |= ICON_DIM;
+
+	if (flags == desired && !force)
+		return;
+
+	WMReplaceInArray(panel->flags, idecks, (void *) (uintptr_t) desired);
 
 	if (!panel->bg && !panel->tile && !selected)
 		WMSetFrameRelief(icon, WRFlat);
 
 	if (image && icon) {
 		RImage *back;
-		int opaq = 255;
+		int opaq = (dim) ? 75 : 255;
 		RImage *tile;
 		WMPoint pos;
 		Pixmap p;
@@ -165,7 +194,7 @@ static void addIconForWindow(WSwitchPanel *panel, WMWidget *parent, WWindow *wwi
 	if (!image)
 		image = get_icon_image(panel->scr, wwin->wm_instance, wwin->wm_class, ICON_TILE_SIZE);
 
-	/* We must resize the icon size (~64) to the switchpanel icon size (~48) */
+	/* We must resize the icon size (~64) to the switch panel icon size (~48) */
 	image = wIconValidateIconSize(image, ICON_SIZE);
 
 	WMAddToArray(panel->images, image);
@@ -177,6 +206,7 @@ static void scrollIcons(WSwitchPanel *panel, int delta)
 	int nfirst = panel->firstVisible + delta;
 	int i;
 	int count = WMGetArrayItemCount(panel->windows);
+	Bool dim;
 
 	if (count <= panel->visibleCount)
 		return;
@@ -193,8 +223,12 @@ static void scrollIcons(WSwitchPanel *panel, int delta)
 
 	panel->firstVisible = nfirst;
 
-	for (i = panel->firstVisible; i < panel->firstVisible + panel->visibleCount; i++)
-		changeImage(panel, i, i == panel->current);
+	for (i = panel->firstVisible; i < panel->firstVisible + panel->visibleCount; i++) {
+		if (i == panel->current)
+			continue;
+		dim = ((int) (uintptr_t) WMGetFromArray(panel->flags, i) & ICON_DIM);
+		changeImage(panel, i, 0, dim, True);
+	}
 }
 
 /*
@@ -204,19 +238,10 @@ static void scrollIcons(WSwitchPanel *panel, int delta)
  */
 static RImage *assemblePuzzleImage(RImage **images, int width, int height)
 {
-	RImage *img = RCreateImage(width, height, 1);
+	RImage *img;
 	RImage *tmp;
 	int tw, th;
 	RColor color;
-	if (!img)
-		return NULL;
-
-	color.red = 0;
-	color.green = 0;
-	color.blue = 0;
-	color.alpha = 255;
-
-	RFillImage(img, &color);
 
 	tw = width - images[0]->width - images[2]->width;
 	th = height - images[0]->height - images[6]->height;
@@ -224,36 +249,40 @@ static RImage *assemblePuzzleImage(RImage **images, int width, int height)
 	if (tw <= 0 || th <= 0)
 		return NULL;
 
+	img = RCreateImage(width, height, 1);
+	if (!img)
+		return NULL;
+
+	color.red   =   0;
+	color.green =   0;
+	color.blue  =   0;
+	color.alpha = 255;
+	RFillImage(img, &color);
+
 	/* top */
-	if (tw > 0) {
-		tmp = RSmoothScaleImage(images[1], tw, images[1]->height);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, 0);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[1], tw, images[1]->height);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, 0);
+	RReleaseImage(tmp);
+
 	/* bottom */
-	if (tw > 0) {
-		tmp = RSmoothScaleImage(images[7], tw, images[7]->height);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[6]->width, height - images[6]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[7], tw, images[7]->height);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[6]->width, height - images[6]->height);
+	RReleaseImage(tmp);
+
 	/* left */
-	if (th > 0) {
-		tmp = RSmoothScaleImage(images[3], images[3]->width, th);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, 0, images[0]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[3], images[3]->width, th);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, 0, images[0]->height);
+	RReleaseImage(tmp);
+
 	/* right */
-	if (th > 0) {
-		tmp = RSmoothScaleImage(images[5], images[5]->width, th);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, width - images[5]->width, images[2]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[5], images[5]->width, th);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, width - images[5]->width, images[2]->height);
+	RReleaseImage(tmp);
+
 	/* center */
-	if (tw > 0 && th > 0) {
-		tmp = RSmoothScaleImage(images[4], tw, th);
-		RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, images[0]->height);
-		RReleaseImage(tmp);
-	}
+	tmp = RSmoothScaleImage(images[4], tw, th);
+	RCopyArea(img, tmp, 0, 0, tmp->width, tmp->height, images[0]->width, images[0]->height);
+	RReleaseImage(tmp);
 
 	/* corners */
 	RCopyArea(img, images[0], 0, 0, images[0]->width, images[0]->height, 0, 0);
@@ -284,7 +313,7 @@ static RImage *getTile(void)
 	return stile;
 }
 
-static void drawTitle(WSwitchPanel *panel, int idecks, char *title)
+static void drawTitle(WSwitchPanel *panel, int idecks, const char *title)
 {
 	char *ntitle;
 	int width = WMWidgetWidth(panel->win);
@@ -328,51 +357,38 @@ static void drawTitle(WSwitchPanel *panel, int idecks, char *title)
 		free(ntitle);
 }
 
-static WMArray *makeWindowListArray(WWindow *curwin, int include_unmapped, Bool class_only)
+static WMArray *makeWindowListArray(WScreen *scr, int include_unmapped, Bool class_only)
 {
 	WMArray *windows = WMCreateArray(10);
-	int fl;
-	WWindow *wwin;
+	WWindow *wwin = scr->focused_window;
 
-	for (fl = 0; fl < 2; fl++) {
-		for (wwin = curwin; wwin; wwin = wwin->prev) {
-			if (((!fl && canReceiveFocus(wwin) > 0) || (fl && canReceiveFocus(wwin) < 0)) &&
-			    (wwin->flags.mapped || include_unmapped)) {
-				if (class_only) {
-					if (!wwin->wm_class || !curwin->wm_class)
-						continue;
-					if (strcmp(wwin->wm_class, curwin->wm_class))
-						continue;
-				}
-
-				if (!WFLAGP(wwin, skip_switchpanel))
-					WMAddToArray(windows, wwin);
-			}
-		}
-		wwin = curwin;
-		/* start over from the beginning of the list */
-		while (wwin->next)
-			wwin = wwin->next;
-
-		for (wwin = curwin; wwin && wwin != curwin; wwin = wwin->prev) {
-			if (((!fl && canReceiveFocus(wwin) > 0) || (fl && canReceiveFocus(wwin) < 0)) &&
-			    (wwin->flags.mapped || include_unmapped)) {
-				if (class_only) {
-				    if (!wwin->wm_class || !curwin->wm_class)
-					continue;
-				    if (strcmp(wwin->wm_class, curwin->wm_class))
+	while (wwin) {
+		if ((canReceiveFocus(wwin) != 0) &&
+		    (wwin->flags.mapped || wwin->flags.shaded || include_unmapped)) {
+			if (class_only)
+				if (!sameWindowClass(scr->focused_window, wwin)) {
+					wwin = wwin->prev;
 					continue;
 				}
-
-				if (!WFLAGP(wwin, skip_switchpanel))
+			if (!WFLAGP(wwin, skip_switchpanel))
 				WMAddToArray(windows, wwin);
-			}
 		}
+		wwin = wwin->prev;
 	}
-
 	return windows;
 }
 
+static WMArray *makeWindowFlagsArray(int count)
+{
+	WMArray *flags = WMCreateArray(count);
+	int i;
+
+	for (i = 0; i < count; i++)
+		WMAddToArray(flags, (void *) 0);
+
+	return flags;
+}
+
 WSwitchPanel *wInitSwitchPanel(WScreen *scr, WWindow *curwin, Bool class_only)
 {
 	WWindow *wwin;
@@ -382,8 +398,10 @@ WSwitchPanel *wInitSwitchPanel(WScreen *scr, WWindow *curwin, Bool class_only)
 	WMRect rect = wGetRectForHead(scr, wGetHeadForPointerLocation(scr));
 
 	panel->scr = scr;
-	panel->windows = makeWindowListArray(curwin, wPreferences.swtileImage != NULL, class_only);
+	panel->windows = makeWindowListArray(scr, wPreferences.swtileImage != NULL, class_only);
 	count = WMGetArrayItemCount(panel->windows);
+	if (count)
+		panel->flags = makeWindowFlagsArray(count);
 
 	if (count == 0) {
 		WMFreeArray(panel->windows);
@@ -468,7 +486,7 @@ WSwitchPanel *wInitSwitchPanel(WScreen *scr, WWindow *curwin, Bool class_only)
 	WMRealizeWidget(panel->win);
 
 	WM_ITERATE_ARRAY(panel->windows, wwin, i) {
-		changeImage(panel, i, 0);
+		changeImage(panel, i, 0, False, True);
 	}
 
 	if (panel->bg) {
@@ -478,8 +496,8 @@ WSwitchPanel *wInitSwitchPanel(WScreen *scr, WWindow *curwin, Bool class_only)
 
 		XSetWindowBackgroundPixmap(dpy, WMWidgetXID(panel->win), pixmap);
 
-#ifdef SHAPE
-		if (mask && wShapeSupported)
+#ifdef USE_XSHAPE
+		if (mask && w_global.xext.shape.supported)
 			XShapeCombineMask(dpy, WMWidgetXID(panel->win), ShapeBounding, 0, 0, mask, ShapeSet);
 #endif
 		if (pixmap)
@@ -498,7 +516,7 @@ WSwitchPanel *wInitSwitchPanel(WScreen *scr, WWindow *curwin, Bool class_only)
 
 	panel->current = WMGetFirstInArray(panel->windows, curwin);
 	if (panel->current >= 0)
-		changeImage(panel, panel->current, 1);
+		changeImage(panel, panel->current, 1, False, False);
 
 	WMMapWidget(panel->win);
 
@@ -514,7 +532,7 @@ void wSwitchPanelDestroy(WSwitchPanel *panel)
 		Window info_win = panel->scr->info_window;
 		XEvent ev;
 		ev.xclient.type = ClientMessage;
-		ev.xclient.message_type = _XA_WM_IGNORE_FOCUS_EVENTS;
+		ev.xclient.message_type = w_global.atom.wm.ignore_focus_events;
 		ev.xclient.format = 32;
 		ev.xclient.data.l[0] = True;
 
@@ -539,6 +557,9 @@ void wSwitchPanelDestroy(WSwitchPanel *panel)
 	if (panel->icons)
 		WMFreeArray(panel->icons);
 
+	if (panel->flags)
+		WMFreeArray(panel->flags);
+
 	WMFreeArray(panel->windows);
 
 	if (panel->tile)
@@ -559,34 +580,54 @@ void wSwitchPanelDestroy(WSwitchPanel *panel)
 	wfree(panel);
 }
 
-WWindow *wSwitchPanelSelectNext(WSwitchPanel *panel, int back, int ignore_minimized)
+WWindow *wSwitchPanelSelectNext(WSwitchPanel *panel, int back, int ignore_minimized, Bool class_only)
 {
-	WWindow *wwin;
+	WWindow *wwin, *curwin, *tmpwin;
 	int count = WMGetArrayItemCount(panel->windows);
 	int orig = panel->current;
+	int i;
+	Bool dim = False;
 
 	if (count == 0)
 		return NULL;
 
-	if (panel->win)
-		changeImage(panel, panel->current, 0);
-
 	if (!wPreferences.cycle_ignore_minimized)
 		ignore_minimized = False;
 
 	if (ignore_minimized && canReceiveFocus(WMGetFromArray(panel->windows, (count + panel->current) % count)) < 0)
 		ignore_minimized = False;
 
+	curwin = WMGetFromArray(panel->windows, orig);
 	do {
-		if (back)
-			panel->current--;
-		else
-			panel->current++;
+		do {
+			if (back)
+				panel->current--;
+			else
+				panel->current++;
 
-		panel->current= (count + panel->current) % count;
-		wwin = WMGetFromArray(panel->windows, panel->current);
+			panel->current = (count + panel->current) % count;
+			wwin = WMGetFromArray(panel->windows, panel->current);
+
+			if (!class_only)
+				break;
+			if (panel->current == orig)
+				break;
+		} while (!sameWindowClass(wwin, curwin));
 	} while (ignore_minimized && panel->current != orig && canReceiveFocus(wwin) < 0);
 
+	WM_ITERATE_ARRAY(panel->windows, tmpwin, i) {
+		if (i == panel->current)
+			continue;
+		if (!class_only || sameWindowClass(tmpwin, curwin))
+			changeImage(panel, i, 0, False, False);
+		else {
+			if (i == orig)
+				dim = True;
+			changeImage(panel, i, 0, True, False);
+		}
+
+	}
+
 	if (panel->current < panel->firstVisible)
 		scrollIcons(panel, panel->current - panel->firstVisible);
 	else if (panel->current - panel->firstVisible >= panel->visibleCount)
@@ -594,8 +635,9 @@ WWindow *wSwitchPanelSelectNext(WSwitchPanel *panel, int back, int ignore_minimi
 
 	if (panel->win) {
 		drawTitle(panel, panel->current, wwin->frame->title);
-
-		changeImage(panel, panel->current, 1);
+		if (panel->current != orig)
+			changeImage(panel, orig, 0, dim, False);
+		changeImage(panel, panel->current, 1, False, False);
 	}
 
 	return wwin;
@@ -605,13 +647,12 @@ WWindow *wSwitchPanelSelectFirst(WSwitchPanel *panel, int back)
 {
 	WWindow *wwin;
 	int count = WMGetArrayItemCount(panel->windows);
+	char *title;
+	int i;
 
 	if (count == 0)
 		return NULL;
 
-	if (panel->win)
-		changeImage(panel, panel->current, 0);
-
 	if (back) {
 		panel->current = count - 1;
 		scrollIcons(panel, count);
@@ -621,10 +662,13 @@ WWindow *wSwitchPanelSelectFirst(WSwitchPanel *panel, int back)
 	}
 
 	wwin = WMGetFromArray(panel->windows, panel->current);
+	title = wwin->frame->title;
 
 	if (panel->win) {
-		drawTitle(panel, panel->current, wwin->frame->title);
-		changeImage(panel, panel->current, 1);
+		for (WMArrayFirst(panel->windows, &(i)); (i) != WANotFound; WMArrayNext(panel->windows, &(i)))
+			changeImage(panel, i, i == panel->current, False, False);
+
+		drawTitle(panel, panel->current, title);
 	}
 
 	return wwin;
@@ -651,8 +695,9 @@ WWindow *wSwitchPanelHandleEvent(WSwitchPanel *panel, XEvent *event)
 	if (focus >= 0 && panel->current != focus) {
 		WWindow *wwin;
 
-		changeImage(panel, panel->current, 0);
-		changeImage(panel, focus, 1);
+		WM_ITERATE_ARRAY(panel->windows, wwin, i) {
+			changeImage(panel, i, i == focus, False, False);
+		}
 		panel->current = focus;
 
 		wwin = WMGetFromArray(panel->windows, focus);
